/*******************************************************************************
***  This file is part of Geometry.                                          ***
***                                                                          ***
***  Copyright (C) 2024                                                      ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Shape.h"
// scanfPoint;
// printfPoint;

#include <limits.h>
// using INT_MIN
#include <stdio.h>
// printf
#include <stdlib.h>
// malloc


const int INVALID_SHAPE_DATA = INT_MIN;

#define PRINT_SHAPE_TYPE(shape)  \
	do  \
	{  \
		const int bytes = printf(" %i - %s\n", shape, #shape);  \
		\
		if (bytes <= 0)  \
			return bytes;  \
	}  \
	while (0)

#define RETURN_SHAPE_TYPE_IF(Type, value)  \
	if (value == (int)Type)  \
		return Type

enum ShapeTypes toShapeTypes(const int weak)
{
	RETURN_SHAPE_TYPE_IF(Circle, weak);
	RETURN_SHAPE_TYPE_IF(Line, weak);
	RETURN_SHAPE_TYPE_IF(Rectangle, weak);
	RETURN_SHAPE_TYPE_IF(Square, weak);
	RETURN_SHAPE_TYPE_IF(Triangle, weak);
	return Unknown;
}

int printfAllShapes(void)
{
	PRINT_SHAPE_TYPE(Circle);
	PRINT_SHAPE_TYPE(Line);
	PRINT_SHAPE_TYPE(Rectangle);
	PRINT_SHAPE_TYPE(Square);
	PRINT_SHAPE_TYPE(Triangle);
	PRINT_SHAPE_TYPE(Unknown);
	return 0;
}

void* mallocShape(enum ShapeTypes shape)
{
	switch (shape)
	{
	case Unknown:
		return NULL;

	case Circle:
		return malloc(sizeof(struct Circle));

	case Line:
		return malloc(sizeof(struct Line));

	case Rectangle:
		return malloc(sizeof(struct Rectangle));

	case Square:
		return malloc(sizeof(struct Square));

	case Triangle:
		return malloc(sizeof(struct Triangle));

	default:
		return NULL;
	};
}

//{ Circle Support

int scanfCircle(struct Circle* shape)
{
	const int readIn = scanfPoint(&shape->centre_);

	if (readIn <= 0)
		return readIn;

	const int radiusReadIn = scanf("%i", &shape->radius_);

	if (radiusReadIn <= 0)
		return radiusReadIn;

	if (shape->radius_ <= 0)
		return INVALID_SHAPE_DATA;

	return readIn + radiusReadIn;
}

int printfCircle(const struct Circle shape)
{
	const int centreBytes = printfPoint(shape.centre_);

	if (centreBytes <= 0)
		return centreBytes;

	const int radiusBytes = printf("r%i", shape.radius_);

	if (radiusBytes <= 0)
		return radiusBytes;

	return centreBytes + radiusBytes;
}

//}

//{ Line Support

int scanfLine(struct Line* shape)
{
	const int readInStart = scanfPoint(&shape->start_);

	if (readInStart <= 0)
		return readInStart;

	const int readInEnd = scanfPoint(&shape->end_);

	if (readInEnd <= 0)
		return readInEnd;

	if (
	        shape->start_.abscissa_.value_ == shape->end_.abscissa_.value_
	        && shape->start_.ordinate_.value_ == shape->end_.ordinate_.value_
	)
		return INVALID_SHAPE_DATA;

	return readInStart + readInEnd;
}

int printfLine(const struct Line shape)
{
	const int startBytes = printfPoint(shape.start_);

	if (startBytes <= 0)
		return startBytes;

	const int endBytes = printfPoint(shape.end_);

	if (endBytes <= 0)
		return endBytes;

	return startBytes + endBytes;
}

//}

//{ Rectangle Support

int scanfRectangle(struct Rectangle* shape)
{
	const int readIn = scanfPoint(&shape->leftEdge_);

	if (readIn <= 0)
		return readIn;

	const int lengthReadIn = scanf("%i", &shape->length_);

	if (lengthReadIn <= 0)
		return lengthReadIn;

	if (shape->length_ <= 0)
		return INVALID_SHAPE_DATA;

	const int widthReadIn = scanf("%i", &shape->width_);

	if (widthReadIn <= 0)
		return widthReadIn;

	if (shape->width_ <= 0)
		return INVALID_SHAPE_DATA;

	return readIn + lengthReadIn + widthReadIn;
}

int printfRectangle(const struct Rectangle shape)
{
	const int leftUpperBytes = printfPoint(shape.leftEdge_);

	if (leftUpperBytes <= 0)
		return leftUpperBytes;

	const int lengthBytes = printf("l%i", shape.length_);

	if (lengthBytes <= 0)
		return lengthBytes;

	const int widthBytes = printf("w%i", shape.width_);

	if (widthBytes <= 0)
		return widthBytes;

	return leftUpperBytes + lengthBytes + widthBytes;
}

//}

//{ Square Support

int scanfSquare(struct Square* shape)
{
	const int readIn = scanfPoint(&shape->leftEdge_);

	if (readIn <= 0)
		return readIn;

	const int lengthReadIn = scanf("%i", &shape->length_);

	if (lengthReadIn <= 0)
		return lengthReadIn;

	if (shape->length_ <= 0)
		return INVALID_SHAPE_DATA;

	return readIn + lengthReadIn;
}

int printfSquare(const struct Square shape)
{
	const int leftUpperBytes = printfPoint(shape.leftEdge_);

	if (leftUpperBytes <= 0)
		return leftUpperBytes;

	const int lengthBytes = printf("l%i", shape.length_);

	if (lengthBytes <= 0)
		return lengthBytes;

	return leftUpperBytes + lengthBytes;
}

//}

//{ Triangle Support

int scanfTriangle(struct Triangle* shape)
{
	const int readInFirst = scanfPoint(&shape->vertex0_);

	if (readInFirst <= 0)
		return readInFirst;

	const int readInSecond = scanfPoint(&shape->vertex1_);

	if (readInSecond <= 0)
		return readInSecond;

	const int readInThird = scanfPoint(&shape->vertex2_);

	if (readInThird <= 0)
		return readInThird;

	const int areaDoubled = shape->vertex0_.abscissa_.value_
	        * (shape->vertex1_.ordinate_.value_ - shape->vertex2_.ordinate_.value_)
	        + shape->vertex1_.abscissa_.value_
	        * (shape->vertex2_.ordinate_.value_ - shape->vertex0_.ordinate_.value_)
	        + shape->vertex2_.abscissa_.value_
	        * (shape->vertex0_.ordinate_.value_ - shape->vertex1_.ordinate_.value_);

	if (areaDoubled == 0)
		return INVALID_SHAPE_DATA;

	return readInFirst + readInSecond + readInThird;
}

int printfTriangle(const struct Triangle shape)
{
	const int firstBytes = printfPoint(shape.vertex0_);

	if (firstBytes <= 0)
		return firstBytes;

	const int secondBytes = printfPoint(shape.vertex1_);

	if (secondBytes <= 0)
		return secondBytes;

	const int thirdBytes = printfPoint(shape.vertex2_);

	if (thirdBytes <= 0)
		return thirdBytes;

	return firstBytes + secondBytes + thirdBytes;
}

//}

#undef RETURN_SHAPE_TYPE_IF
#undef PRINT_SHAPE_TYPE
